
#!/usr/bin/env python3

"""
  Course:      Data Structures and Algorithms for CL III - WS2425
  Assignment:  (Assignment number)
  Author:      (Your name)
  Description: (...)
  Honor Code:  I/We pledge that this program represents my/our own work.
"""

import sys, argparse, os


class FSA:
    """ A class representing finite state automata.
    Args:
        deterministic: The automaton is deterministic
    Attributes:
        transitions: transitions kept as a dictionary
            where keys are the tuple (source_state, symbol),
            values are the target state for DFA
            and a list of target states for NFA.
            Note that we do not require a dedicated 'sink' state.
            Any undefined transition should cause the FSA to reject the
            string immediately.
        start_state: number/name of the start state
        accepting: the set of accepting states
        is_deterministic (boolean): whether the FSA is deterministic or not
    """

    def __init__(self, deterministic=True):
        self._reset(deterministic)

    def _reset(self, deterministic):
        self.start_state = None
        self.is_deterministic = deterministic
        self.transitions = dict()
        self.accepting = set()
        self._alphabet = set()  # just for convenience, we can
        self._states = set()  # always read it off from transitions

    def add_arc(self, s1, sym, s2=None, accepting=False):
        """ Add an arc from state s1 to s2 with symbol.

        If s2 is not specified, the method creates a new state.
        The return value is s2.

        Note that we follow a convention to specify the states by
        integers. However, the method should work with other ways of
        specifying the state names/numbers.
        """
        if self.start_state is None:
            self.start_state = s1
            self._states.add(s1)
        if s2 is None:
            s2 = len(self._states)
            while s2 in self._states: s2 += 1
        self._states.add(s2)
        self._alphabet.add(sym)
        if (s1, sym) not in self.transitions:
            self.transitions[(s1, sym)] = set()
        self.transitions[(s1, sym)].add(s2)
        if accepting:
            self.accepting.add(s2)
        if len(self.transitions[(s1, sym)]) > 1:
            self.is_deterministic = False
        return s2

    def mark_accept(self, state):
        self.accepting.add(state)

    def unmark_accept(self, state):
        if state in self.accepting:
            self.accepting.remove(state)

    def toggle_accept(self, state):
        if state in self.accepting:
            self.accepting.remove(state)
        else:
            self.accepting.add(state)

    def is_accepting(self, state):
        return state in self.accepting

    def move(self, sym, s1=None):
        """ Return the state(s) reachable from 's1' on 'symbol'
        """
        if s1 is None:
            s1 = self.start_state
        if (s1, sym) not in self.transitions:
            return None
        else:
            return self.transitions[(s1, sym)]

    def _recognize_dfa(self, s):
        """ DFA recognition of 's'.
        """
        state = self.start_state
        for sym in s:
            states = self.transitions.get((state, sym), None)
            if states is None:
                return False
            else:
                assert isinstance(states, set)
                state = next(iter(states))
        if state in self.accepting:
            return True
        else:
            return False

    def _recognize_nfa(self, s):
        """ NFA recognition of 's' using a stack-based agenda.
        """
        agenda = []
        # state = self.start_state
        inp_pos = 0
        for node in self.transitions.get((self.start_state, s[inp_pos]), []):
            agenda.append((node, inp_pos + 1))
        while agenda:
            node, inp_pos = agenda.pop()
            if inp_pos == len(s):
                if node in self.accepting:
                    return True
            else:
                for node in self.transitions.get((node, s[inp_pos]), []):
                    agenda.append((node, inp_pos + 1))
        return False

    def recognize(self, s):
        """ Recognize the given string 's', return a boolean value
        """
        if self.is_deterministic:
            return self._recognize_dfa(s)
        else:
            return self._recognize_nfa(s)

    def write_att(self, filename=None, dot=False):
        """ Write the FSA to an AT&T formatted text file.

        Not required, but may be useful for testing.
        If the 'dot' argument is True, writes a '.dot' file rather
        than an AT&T file.
        """
        if filename:
            fp = open(filename, 'w', encoding="utf8")
        else:
            fp = sys.stdout
        if dot:
            print("digraph {\n"
                  "  rankdir = LR;\n"
                  "  start[style=invis];\n"
                  "  node[shape=circle];", file=fp)
            fmt_transition = "  {} -> {} [label=\"{}\"];"
            fmt_accepting = "  {} [shape=doublecircle];"
        else:
            fmt_transition = "{}\t{}\t{}"
            fmt_accepting = "{}"

        # make sure the first symbol to output is the start state
        for sym in self._alphabet:
            if (self.start_state, sym) in self.transitions:
                for s2 in self.transitions[(self.start_state, sym)]:
                    print(fmt_transition.format(self.start_state, s2, sym),
                          file=fp)
        for s1, sym in self.transitions:
            if s1 != self.start_state:
                for s2 in self.transitions[(s1, sym)]:
                    print(fmt_transition.format(s1, s2, sym), file=fp)
        for st in self.accepting:
            print(fmt_accepting.format(st), file=fp)
        if dot:
            print("}", file=fp)
        if filename:
            fp.close()

    def to_dot(self, filename=None):
        """ Build a string that contains the 'dot' representation
        of the graph, and write to 'filename', or standard output
        if not specified.

        You may use this function to visualize your results and
        visualize the changes you perform on the FSA.
        The functionality is implemented in write_att(),
        this method is just a wrapper.
        """
        self.write_att(filename, dot=True)

    def insert_words(self, words):
        """Reset the FSA and insert given words.

        The result of this should be trie. We will obtain a more compact
        automaton by FSA minimization.
        """
        # TODO  Exercise 1
        pass

    def minimize(self, verbose=False):
        """Minimize the automaton.

        If verbose=True, print out the number of states and arcs
        before and after the minimization.
        """
        # TODO  Exercise 2
        pass


if __name__ == '__main__':
    argp = argparse.ArgumentParser()
    argp.add_argument('input', help="Input file, from which to read the lexicon.")
    argp.add_argument('output', help="Folder, where to save the FSA.")
    argp.add_argument('--compact', action='store_true', help="Compact the lexicon by minimizing the FSA.")
    args = argp.parse_args()

    # read the lexicon
    words = set()
    with open(args.input, 'rt', encoding="utf8") as f:
        for line in f:
            words.add(line.strip())

    # setup fsa
    m = FSA()
    m.insert_words(words)

    # minimize, if specified
    if args.compact:
        m.minimize()

    # give it a folder, can be non-existing
    if not os.path.isdir(args.output):
        os.mkdir(args.output)

    # and save
    m.write_att(filename=f'{args.output}/automata_att.att', dot=False)
    m.to_dot(filename=f'{args.output}/automata_dot.dot')

    print(f'your automaton was saved at \'{os.path.abspath(args.output)}\'!')
